package expo.modules.browserview

import expo.modules.core.errors.CurrentActivityNotFoundException
import android.content.Intent
import android.net.Uri
import android.os.Bundle
import android.text.TextUtils
import androidx.browser.customtabs.CustomTabsIntent
import androidx.core.os.bundleOf
import expo.modules.core.utilities.ifNull
import expo.modules.kotlin.modules.Module
import expo.modules.kotlin.modules.ModuleDefinition

private const val SERVICE_PACKAGE_KEY = "servicePackage"
private const val BROWSER_PACKAGES_KEY = "browserPackages"
private const val SERVICE_PACKAGES_KEY = "servicePackages"
private const val PREFERRED_BROWSER_PACKAGE = "preferredBrowserPackage"
private const val DEFAULT_BROWSER_PACKAGE = "defaultBrowserPackage"

private const val MODULE_NAME = "PhantomBrowserView"

class BrowserViewModule : Module() {
    override fun definition() = ModuleDefinition {
        Name(MODULE_NAME)

        OnCreate {
            customTabsResolver = CustomTabsActivitiesHelper(appContext.activityProvider)
            connectionHelper = CustomTabsConnectionHelper(
                requireNotNull(appContext.reactContext) {
                    "Cannot initialize BrowserView, ReactContext is null"
                }
            )
        }

        OnActivityDestroys {
            connectionHelper.destroy()
        }

        AsyncFunction("warmUpAsync") { packageName: String? ->
            val resolvedPackageName = givenOrPreferredPackageName(packageName)
            connectionHelper.warmUp(resolvedPackageName)
            return@AsyncFunction bundleOf(
                SERVICE_PACKAGE_KEY to resolvedPackageName
            )
        }

        AsyncFunction("coolDownAsync") { packageName: String? ->
            val resolvedPackageName = givenOrPreferredPackageName(packageName)
            val result = if (connectionHelper.coolDown(resolvedPackageName)) {
                bundleOf(
                    SERVICE_PACKAGE_KEY to resolvedPackageName
                )
            } else {
                Bundle()
            }
            return@AsyncFunction result
        }

        AsyncFunction("mayInitWithUrlAsync") { url: String, packageName: String? ->
            val resolvedPackageName = givenOrPreferredPackageName(packageName)
            connectionHelper.mayInitWithUrl(resolvedPackageName, Uri.parse(url))
            return@AsyncFunction bundleOf(
                SERVICE_PACKAGE_KEY to resolvedPackageName
            )
        }

        // throws CurrentActivityNotFoundException
        AsyncFunction<Bundle>("getCustomTabsSupportingBrowsersAsync") {
            val activities = customTabsResolver.customTabsResolvingActivities
            val services = customTabsResolver.customTabsResolvingServices
            val preferredPackage =
                customTabsResolver.getPreferredCustomTabsResolvingActivity(activities)
            val defaultPackage = customTabsResolver.defaultCustomTabsResolvingActivity

            // It might happen, that default activity does not support Chrome Tabs. Then it will be ResolvingActivity and we don't want to return it as a result.
            val defaultCustomTabsPackage: String? =
                defaultPackage.takeIf { activities.contains(it) }

            return@AsyncFunction Bundle().apply {
                putStringArrayList(BROWSER_PACKAGES_KEY, activities)
                putStringArrayList(SERVICE_PACKAGES_KEY, services)
                putString(PREFERRED_BROWSER_PACKAGE, preferredPackage)
                putString(DEFAULT_BROWSER_PACKAGE, defaultCustomTabsPackage)
            }
        }

        // throws CurrentActivityNotFoundException
        AsyncFunction("openBrowserAsync") { url: String, options: OpenBrowserOptions ->
            val ctIntent = createCustomTabsIntent(options)

            val intent = ctIntent.intent

            intent.data = Uri.parse(url)

            if (!customTabsResolver.canResolveIntent(intent)) {
                throw NoMatchingActivityException()
            }

            // Establish service connection for partial custom tabs
            val packageName = options.browserPackage
                ?: customTabsResolver.getPreferredCustomTabsResolvingActivity(null)
            if (packageName != null) {
                connectionHelper.warmUp(packageName)
            }

            appContext.reactContext?.let { ctIntent.launchUrl(it, Uri.parse(url)) }
//            customTabsResolver.startCustomTabs(intent)

            return@AsyncFunction bundleOf(
                "type" to "opened"
            )
        }
    }

    // these must be `internal` to be able to be injected in tests
    internal lateinit var customTabsResolver: CustomTabsActivitiesHelper
    internal lateinit var connectionHelper: CustomTabsConnectionHelper

    private fun createCustomTabsIntent(options: OpenBrowserOptions): CustomTabsIntent {
        // Get a session for the custom tabs
        val session = connectionHelper.getSession()
        val builder = if (session != null) {
            CustomTabsIntent.Builder(session)
        } else {
            CustomTabsIntent.Builder()
        }

        val color = options.toolbarColor
        if (color != null) {
            builder.setToolbarColor(color)
        }

        val secondaryColor = options.secondaryToolbarColor
        if (secondaryColor != null) {
            builder.setSecondaryToolbarColor(secondaryColor)
        }

        builder.setShowTitle(options.showTitle)

        if (options.enableDefaultShareMenuItem) {
            builder.addDefaultShareMenuItem()
        }

        val displayMetrics = appContext.reactContext?.resources?.displayMetrics
        val screenHeight = displayMetrics?.heightPixels ?: 0
        val initialHeight = if (options.halfSheet) {
            (screenHeight * 0.3).toInt()
        } else {
            (screenHeight * 1.0).toInt()
        }

        // Set both with builder method and with legacy extras for better compatibility
        builder.setInitialActivityHeightPx(
            initialHeight,
            CustomTabsIntent.ACTIVITY_HEIGHT_ADJUSTABLE
        )

        builder.setToolbarCornerRadiusDp(10)

        // Disable animations by setting them to 0
        // This prevents the "reappear and animate down" visual glitch
        builder.setStartAnimations(appContext.reactContext!!, 0, 0)
        builder.setExitAnimations(appContext.reactContext!!, 0, 0)

        val customTabIntent = builder.build()

// We cannot use the builder's method enableUrlBarHiding, because there is
        // no corresponding disable method and some browsers enable it by default.
        customTabIntent.intent.putExtra(
            CustomTabsIntent.EXTRA_ENABLE_URLBAR_HIDING,
            options.enableBarCollapsing
        )


        val packageName = options.browserPackage
        if (!TextUtils.isEmpty(packageName)) {
            customTabIntent.intent.setPackage(packageName)
        }

        if (options.shouldCreateTask) {
            customTabIntent.intent.addFlags(Intent.FLAG_ACTIVITY_NEW_TASK)

            if (!options.showInRecents) {
                customTabIntent.intent.addFlags(Intent.FLAG_ACTIVITY_EXCLUDE_FROM_RECENTS)
                customTabIntent.intent.addFlags(Intent.FLAG_ACTIVITY_NO_HISTORY)
            }
        }


        return customTabIntent
    }

    /**
     * @throws NoPreferredPackageFound
     */
    private fun givenOrPreferredPackageName(packageName: String?): String {
        val resolvedPackageName: String? = try {
            packageName?.takeIf { it.isNotEmpty() }.ifNull {
                customTabsResolver.getPreferredCustomTabsResolvingActivity(null)
            }
        } catch (ex: CurrentActivityNotFoundException) {
            throw NoPreferredPackageFound()
        } catch (ex: PackageManagerNotFoundException) {
            throw NoPreferredPackageFound()
        }

        return resolvedPackageName?.takeIf { it.isNotEmpty() }
            ?: throw NoPreferredPackageFound()
    }
}
