// Copyright 2022-present 650 Industries. All rights reserved.

import SafariServices
import ExpoModulesCore
import AuthenticationServices

final public class BrowserViewModule: Module {
  private var currentBrowserViewSession: BrowserViewSession?
  private var currentAuthSession: WebAuthSession?
  private var vcDidPresent = false

  private func isValid(url: URL) -> Bool {
    return url.scheme == "http" || url.scheme == "https"
  }

  public func definition() -> ModuleDefinition {
    Name("PhantomBrowserView")

    AsyncFunction("openBrowserAsync") { (url: URL, options: BrowserViewOptions, promise: Promise) in
      if vcDidPresent {
        self.currentBrowserViewSession = nil
        vcDidPresent = false
      }

      guard self.currentBrowserViewSession == nil else {
        throw BrowserViewAlreadyOpenException()
      }

      guard self.isValid(url: url) else {
        throw BrowserViewInvalidURLException()
      }

      self.currentBrowserViewSession = BrowserViewSession(url: url, options: options) { [promise] type in
        promise.resolve(["type": type])
        self.currentBrowserViewSession = nil
      } didPresent: {
        self.vcDidPresent = true
      }

      self.currentBrowserViewSession?.open()
    }
    .runOnQueue(.main)

    AsyncFunction("dismissBrowser") {
      currentBrowserViewSession?.dismiss()
      currentBrowserViewSession = nil
    }
    .runOnQueue(.main)

    // MARK: - AuthSession

    AsyncFunction("openAuthSessionAsync") { (authUrl: URL, redirectUrl: URL?, options: AuthSessionOptions, promise: Promise) throws in
      guard self.currentAuthSession?.isOpen != true else {
        throw BrowserViewAlreadyOpenException()
      }
      self.currentAuthSession = WebAuthSession(authUrl: authUrl, redirectUrl: redirectUrl, options: options)
      self.currentAuthSession?.open(promise)
    }
    .runOnQueue(.main)

    AsyncFunction("dismissAuthSession") {
      self.currentAuthSession?.dismiss()
      self.currentAuthSession = nil
    }
    .runOnQueue(.main)

    // MARK: - Stubs for jest-expo-mock-generator

    AsyncFunction("warmUpAsync") {}
    AsyncFunction("coolDownAsync") {}
    AsyncFunction("mayInitWithUrlAsync") {}
    AsyncFunction("getCustomTabsSupportingBrowsers") {}
  }
}
